@@ -1653,104 +1653,6 @@ class TestAttachments:
 
 class TestFeatureServiceOverwriter:
 
-    def test_rename_columns_from_agol_handles_special_and_space(self):
-        cols = ['Test Name:']
-
-        renamed = palletjack.FeatureServiceOverwriter._rename_columns_for_agol(cols)
-
-        assert renamed == {'Test Name:': 'Test_Name_'}
-
-    def test_rename_columns_from_agol_handles_special_and_space_leaves_others_alone(self):
-        cols = ['Test Name:', 'FooName']
-
-        renamed = palletjack.FeatureServiceOverwriter._rename_columns_for_agol(cols)
-
-        assert renamed == {'Test Name:': 'Test_Name_', 'FooName': 'FooName'}
-
-    def test_rename_columns_from_agol_retains_underscores(self):
-        cols = ['Test Name:', 'Foo_Name']
-
-        renamed = palletjack.FeatureServiceOverwriter._rename_columns_for_agol(cols)
-
-        assert renamed == {'Test Name:': 'Test_Name_', 'Foo_Name': 'Foo_Name'}
-
-    def test_check_fields_match_normal(self, mocker):
-        mock_fl = mocker.Mock()
-        mock_fl.properties = {
-            'fields': [
-                {
-                    'name': 'Foo'
-                },
-                {
-                    'name': 'Bar'
-                },
-            ]
-        }
-        df = pd.DataFrame(columns=['Foo', 'Bar'])
-        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
-
-        palletjack.FeatureServiceOverwriter._check_fields_match(overwriter, mock_fl, df)
-
-    def test_check_fields_match_raises_error_on_extra_new_field(self, mocker):
-        mock_fl = mocker.Mock()
-        mock_fl.properties = {
-            'fields': [
-                {
-                    'name': 'Foo'
-                },
-                {
-                    'name': 'Bar'
-                },
-            ]
-        }
-        df = pd.DataFrame(columns=['Foo', 'Bar', 'Baz'])
-        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
-
-        with pytest.raises(RuntimeError) as exc_info:
-            palletjack.FeatureServiceOverwriter._check_fields_match(overwriter, mock_fl, df)
-
-        assert exc_info.value.args[
-            0] == 'New dataset contains the following fields that are not present in the live dataset: {\'Baz\'}'
-
-    def test_check_fields_match_ignores_new_shape_field(self, mocker):
-        mock_fl = mocker.Mock()
-        mock_fl.properties = {
-            'fields': [
-                {
-                    'name': 'Foo'
-                },
-                {
-                    'name': 'Bar'
-                },
-            ]
-        }
-        df = pd.DataFrame(columns=['Foo', 'Bar', 'SHAPE'])
-        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
-
-        palletjack.FeatureServiceOverwriter._check_fields_match(overwriter, mock_fl, df)
-
-    def test_check_fields_match_warns_on_missing_new_field(self, mocker, caplog):
-        mock_fl = mocker.Mock()
-        mock_fl.properties = {
-            'fields': [
-                {
-                    'name': 'Foo'
-                },
-                {
-                    'name': 'Bar'
-                },
-                {
-                    'name': 'Baz'
-                },
-            ]
-        }
-        df = pd.DataFrame(columns=['Foo', 'Bar'])
-        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
-
-        palletjack.FeatureServiceOverwriter._check_fields_match(overwriter, mock_fl, df)
-
-        assert 'New dataset does not contain the following fields that are present in the live dataset: {\'Baz\'}' in caplog.text
-
     def test_truncate_existing_data_normal(self, mocker):
         fl_mock = mocker.Mock()
         fl_mock.manager.truncate.return_value = {
@@ -1789,86 +1691,6 @@ class TestFeatureServiceOverwriter:
 
         palletjack.FeatureServiceOverwriter._truncate_existing_data(overwriter, fl_mock, 0, 'abc')
 
-    def test_retry_returns_on_first_success(self, mocker):
-        mock = mocker.Mock()
-        mock.function.return_value = 42
-
-        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
-
-        answer = overwriter._retry(mock.function, 'a', 'b')
-
-        assert answer == 42
-        assert mock.function.call_count == 1
-
-    def test_retry_returns_after_one_failure(self, mocker):
-        mock = mocker.Mock()
-        mock.function.side_effect = [urllib.error.HTTPError('a', 'b', 'c', 'd', 'e'), 42]
-        mocker.patch('palletjack.updaters.sleep')
-
-        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
-
-        answer = overwriter._retry(mock.function, 'a', 'b')
-
-        assert answer == 42
-        assert mock.function.call_count == 2
-
-    def test_retry_returns_after_two_failures(self, mocker):
-        mock = mocker.Mock()
-        mock.function.side_effect = [
-            urllib.error.HTTPError('a', 'b', 'c', 'd', 'e'),
-            urllib.error.HTTPError('a', 'b', 'c', 'd', 'e'),
-            42,
-        ]
-        mocker.patch('palletjack.updaters.sleep')
-
-        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
-
-        answer = overwriter._retry(mock.function, 'a', 'b')
-
-        assert answer == 42
-        assert mock.function.call_count == 3
-
-    def test_retry_fails_after_four_failures(self, mocker):
-        mock = mocker.Mock()
-        mock.function.side_effect = [
-            urllib.error.HTTPError('a', 'b', 'c', 'd', 'e'),
-            urllib.error.HTTPError('a', 'b', 'c', 'd', 'e'),
-            urllib.error.HTTPError('a', 'b', 'c', 'd', 'e'),
-            urllib.error.HTTPError('a', 'b', 'c', 'd', 'e'),
-            42,
-        ]
-        mocker.patch('palletjack.updaters.sleep')
-
-        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
-
-        with pytest.raises(urllib.error.HTTPError):
-            answer = overwriter._retry(mock.function, 'a', 'b')
-
-        # assert answer == 42
-        assert mock.function.call_count == 4
-
-    def test_retry_retries_on_request_http_error(self, mocker):
-        mock = mocker.Mock()
-        mock.function.side_effect = [requests.exceptions.HTTPError, 42]
-        mocker.patch('palletjack.updaters.sleep')
-
-        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
-
-        answer = overwriter._retry(mock.function, 'a', 'b')
-
-        assert answer == 42
-
-    def test_retry_retries_on_request_ssl_error(self, mocker):
-        mock = mocker.Mock()
-        mock.function.side_effect = [requests.exceptions.SSLError, 42]
-        mocker.patch('palletjack.updaters.sleep')
-
-        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
-
-        answer = overwriter._retry(mock.function, 'a', 'b')
-
-        assert answer == 42
-
     def test_append_new_data_doesnt_raise_on_normal(self, mocker):
         mock_df = mocker.Mock()
         mock_fl = mocker.Mock()
@@ -2007,51 +1829,6 @@ class TestFeatureServiceOverwriter:
         assert 'Append failed; attempting to re-load truncated data...' in caplog.text
         assert 'features reloaded' not in caplog.text
 
-    def test_replace_nan_series_with_empty_strings_one_empty_one_non_empty_float(self, mocker):
-        df = pd.DataFrame({
-            'normal': [1., 2., 3.],
-            'empty': [np.nan, np.nan, np.nan],
-        })
-
-        fixed_df = palletjack.FeatureServiceOverwriter._replace_nan_series_with_empty_strings(mocker.Mock(), df)
-
-        test_df = pd.DataFrame({
-            'normal': [1., 2., 3.],
-            'empty': ['', '', ''],
-        })
-
-        tm.assert_frame_equal(fixed_df, test_df)
-
-    def test_replace_nan_series_with_empty_strings_other_series_has_nan(self, mocker):
-        df = pd.DataFrame({
-            'normal': [1., 2., np.nan],
-            'empty': [np.nan, np.nan, np.nan],
-        })
-
-        fixed_df = palletjack.FeatureServiceOverwriter._replace_nan_series_with_empty_strings(mocker.Mock(), df)
-
-        test_df = pd.DataFrame({
-            'normal': [1., 2., np.nan],
-            'empty': ['', '', ''],
-        })
-
-        tm.assert_frame_equal(fixed_df, test_df)
-
-    def test_replace_nan_series_with_empty_strings_other_series_is_empty_str(self, mocker):
-        df = pd.DataFrame({
-            'normal': ['', '', ''],
-            'empty': [np.nan, np.nan, np.nan],
-        })
-
-        fixed_df = palletjack.FeatureServiceOverwriter._replace_nan_series_with_empty_strings(mocker.Mock(), df)
-
-        test_df = pd.DataFrame({
-            'normal': ['', '', ''],
-            'empty': ['', '', ''],
-        })
-
-        tm.assert_frame_equal(fixed_df, test_df)
-
     def test_save_truncated_data_calls_write_with_json(self, mocker):
         mock_df = pd.DataFrame({
             'foo': [1],
@@ -2105,3 +1882,146 @@ class TestFeatureServiceOverwriter:
         out_path = palletjack.FeatureServiceOverwriter._save_truncated_data(mocker.Mock(), mock_sdf, 'foo')
 
         assert out_path == Path('foo/old_data_foo-date.json')
+
+
+class TestFeatureServiceInlineUpdaterUpsert:
+
+    def test_upsert_new_data_doesnt_raise_on_normal(self, mocker):
+        mock_df = mocker.Mock()
+        mock_fl = mocker.Mock()
+        mock_fl.append.return_value = (True, {'message': 'foo'})
+        mocker.patch('palletjack.utils.rename_columns_for_agol')
+        updater = palletjack.FeatureServiceInlineUpdater(mocker.Mock(), mocker.Mock(), 'foo')
+
+        updater._upsert_new_data(mock_fl, mock_df, 'abc', 0)
+
+    def test_upsert_new_data_retries_on_httperror(self, mocker):
+        mock_df = mocker.Mock()
+        mock_fl = mocker.Mock()
+        mock_fl.append.side_effect = [Exception, (True, {'message': 'foo'})]
+        mocker.patch('palletjack.utils.rename_columns_for_agol')
+
+        updater = palletjack.FeatureServiceInlineUpdater(mocker.Mock(), mocker.Mock(), 'foo')
+        updater._upsert_new_data(mock_fl, mock_df, 'abc', 0)
+
+    def test_upsert_new_data_raises_on_False_result(self, mocker):
+        mock_df = mocker.Mock()
+        mock_fl = mocker.Mock()
+        mock_fl.append.return_value = (False, {'message': 'foo'})
+        mocker.patch('palletjack.utils.rename_columns_for_agol')
+
+        updater = palletjack.FeatureServiceInlineUpdater(mocker.Mock(), mocker.Mock(), 'foo')
+
+        with pytest.raises(RuntimeError) as exc_info:
+            updater._upsert_new_data(mock_fl, mock_df, 'abc', 0)
+
+        assert exc_info.value.args[
+            0] == 'Failed to append data to layer id 0 in itemid abc. Append should have been rolled back.'
+
+    def test_upsert_new_data_in_hosted_feature_layer_normal(self, mocker):
+        mock_fl = mocker.Mock()
+        # mock_fl.manager.truncate.return_value = {
+        #     'submissionTime': 123,
+        #     'lastUpdatedTime': 124,
+        #     'status': 'Completed',
+        # }
+        mock_fl.properties = {
+            'fields': [
+                {
+                    'name': 'Foo'
+                },
+                {
+                    'name': 'Bar'
+                },
+            ],
+            'indexes': [
+                {
+                    'fields': 'Foo',
+                    'isUnique': True
+                },
+            ]
+        }
+
+        mock_fl.append.return_value = (True, {'recordCount': 42})
+
+        fl_class_mock = mocker.Mock()
+        fl_class_mock.fromitem.return_value = mock_fl
+        mocker.patch('arcgis.features.FeatureLayer', fl_class_mock)
+
+        new_dataframe = pd.DataFrame(columns=['Foo', 'Bar'])
+        mocker.patch.object(pd.DataFrame, 'spatial')
+
+        updater = palletjack.FeatureServiceInlineUpdater(mocker.Mock(), new_dataframe, 'Foo')
+
+        uploaded_features = updater.upsert_new_data_in_hosted_feature_layer('abc')
+
+        assert uploaded_features == 42
+
+    def test_upsert_new_data_in_hosted_feature_layer_handles_field_renaming(self, mocker):
+        mock_fl = mocker.Mock()
+        # mock_fl.manager.truncate.return_value = {
+        #     'submissionTime': 123,
+        #     'lastUpdatedTime': 124,
+        #     'status': 'Completed',
+        # }
+        mock_fl.properties = {
+            'fields': [
+                {
+                    'name': 'Foo_field'
+                },
+                {
+                    'name': 'Bar'
+                },
+            ],
+            'indexes': [
+                {
+                    'fields': 'Bar',
+                    'isUnique': True
+                },
+            ]
+        }
+        mock_fl.append.return_value = (True, {'recordCount': 42})
+
+        fl_class_mock = mocker.Mock()
+        fl_class_mock.fromitem.return_value = mock_fl
+        mocker.patch('arcgis.features.FeatureLayer', fl_class_mock)
+
+        new_dataframe = pd.DataFrame(columns=['Foo field', 'Bar'])
+        mocker.patch.object(pd.DataFrame, 'spatial')
+
+        updater = palletjack.FeatureServiceInlineUpdater(mocker.Mock(), new_dataframe, 'Bar')
+
+        uploaded_features = updater.upsert_new_data_in_hosted_feature_layer('abc')
+
+        assert uploaded_features == 42
+
+
+class TestFeatureServiceInlineUpdaterInit:
+
+    def test_init_raises_on_missing_index_field(self, mocker):
+
+        new_dataframe = pd.DataFrame(columns=['Foo_field', 'Bar'])
+        # mocker.patch.object(pd.DataFrame, 'spatial')
+
+        with pytest.raises(KeyError) as exc_info:
+            updater = palletjack.FeatureServiceInlineUpdater(mocker.Mock(), new_dataframe, 'Baz')
+
+        assert exc_info.value.args[0] == 'Index column Baz not found in dataframe columns'
+
+    def test_init_renames_dataframe_columns(self, mocker):
+
+        new_dataframe = pd.DataFrame(columns=['Foo field', 'Bar'])
+        # mocker.patch.object(pd.DataFrame, 'spatial')
+
+        updater = palletjack.FeatureServiceInlineUpdater(mocker.Mock(), new_dataframe, 'Bar')
+
+        assert list(updater.new_dataframe.columns) == ['Foo_field', 'Bar']
+
+    def test_init_renames_index_column(self, mocker):
+
+        new_dataframe = pd.DataFrame(columns=['Foo field', 'Bar'])
+        # mocker.patch.object(pd.DataFrame, 'spatial')
+
+        updater = palletjack.FeatureServiceInlineUpdater(mocker.Mock(), new_dataframe, 'Foo field')
+
+        assert updater.index_column == 'Foo_field'
